import configparser
import glob
import logging
import os
import xml.etree.ElementTree as ET
from collections import defaultdict
from pathlib import Path

import pandas as pd
import pkg_resources
from jinja2 import Environment, FileSystemLoader, PackageLoader

from mystran_validation import acopy, cleandir, get_conf
from mystran_validation.utils import slugify, working_directory

try:
    loader = PackageLoader("mystran_validation", "data/html_template.j2")
except ValueError:
    # probably installed using `pip install -e .`
    logging.warning("logging from FileSystem")
    tpl_folder = (pkg_resources.resource_filename("mystran_validation", "data"),)
    logging.warning(f"{tpl_folder=}")
    loader = FileSystemLoader(tpl_folder)
env = Environment(loader=loader)
env.filters["slugify"] = slugify


class TestSuite:
    """TestSuite: container for all testsuite test cases. Normally, this is only one."""

    def __init__(self, xml):
        self.meta = xml.attrib
        self.testcases = {}
        # ---------------------------------------------------------------------
        # configure XML / HTML / assets dir
        config_fpath, config = get_conf()
        for i, tc in enumerate(xml.findall("./testcase")):
            tcobject = TestCase(tc)
            if tcobject.classname not in self.testcases:
                tcparent = TestCaseParent(tcobject.classname)
                self.testcases[tcobject.classname] = tcparent
            self.testcases[tcobject.classname].add_test_case(tcobject)

    def __repr__(self):
        return "testsuite errors={errors} failures={failures} skipped={skipped} tests={tests} time={time} timestamp={timestamp}".format(
            **self.meta
        )


class TestCaseParent:
    """thin wrapper equivalent to ini file"""

    def __init__(self, classname):
        self.classname = classname
        self.testcases = []
        self.meta = {}
        self._status = defaultdict(int)
        self.marks = set()
        self.assets = defaultdict(dict)
        self.rel_assets = defaultdict(dict)
        self.test_config = configparser.ConfigParser()
        self.rootdir = None

    def get_assets(self, tcobject, section):
        cfg = dict(self.test_config[section].items())
        if self.rootdir is None:
            self.rootdir = Path(tcobject.properties["test-config"]).parent
        if section == "DEFAULT":
            neutral_output = tcobject.properties.get("output")
            if neutral_output:
                cfg["neutral output"] = neutral_output
            f06_output = tcobject.properties.get("output_f06")
            if f06_output:
                cfg["F06 output"] = f06_output
            # also collect user defined assets
            if "__ASSETS__" in self.test_config.sections():
                assets_cfg = dict(self.test_config["__ASSETS__"].items())
                for k, v in assets_cfg.items():
                    if cfg.get(k) == v:
                        continue
                    if (",") in v:
                        path, key = [t.strip() for t in v.split(",")]
                    else:
                        path = v
                        key = " ".join([token.title() for token in k.split("_")])
                    self.assets[section][key] = self.rootdir / Path(path)
        else:
            # get only keys differing from "DEFAULT"
            default_cfg = dict(self.test_config[section].items())
            cfg = {k: v for k, v in cfg.items() if v != default_cfg.get(k)}
            xlsxdiff = tcobject.properties.get("xlsxdiff")
            if xlsxdiff and Path(xlsxdiff).exists():
                self.assets[section]["xlsxdiff"] = Path(xlsxdiff)
        for k in ("bulk", "reference", "neutral output", "F06 output"):
            if k in cfg:
                path = self.rootdir / Path(cfg[k])
                if path.exists():
                    self.assets[section][k] = path

    def copy_assets(self, section):
        """copy files and store assets as path relative to HTML file"""
        html_outdir = Path(os.environ["MYSTRAN_VALIDATION_BUILDDIR"])
        assets_outdir = html_outdir / "assets" / Path(self.classname).stem / section
        cleandir(assets_outdir, parents=True)
        assets = dict(self.assets).get(section, {})
        if "xlsxdiff" in assets:
            if "xlsxdiff" not in self.rel_assets["DEFAULT"]:
                target = assets_outdir.parent / "DEFAULT" / assets["xlsxdiff"].name
                acopy(assets["xlsxdiff"], target)
                self.rel_assets["DEFAULT"]["xlsxdiff"] = target.relative_to(html_outdir)
            assets.pop("xlsxdiff")
        for k, src in assets.items():
            # if k == "xlsxdiff":
            #     breakpoint()
            target = assets_outdir / src.name
            acopy(src, target)
            target = target.relative_to(html_outdir)
            self.rel_assets[section][k] = target

    def __hash__(self):
        return hash(self.classname)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.classname == other.classname

    def add_test_case(self, tcobject):
        if not self.testcases:
            # actions to perform once only
            self.test_config.read(tcobject.properties["test-config"])
            self.get_assets(tcobject, section="DEFAULT")
            self.shortdesc = tcobject.properties.pop("shortdesc", "")
            self.copy_assets("DEFAULT")
        self.get_assets(tcobject, section=tcobject.name)
        self.copy_assets(section=tcobject.name)
        self.testcases.append(tcobject)
        self._status[tcobject.status] += 1
        self.marks |= eval(tcobject.properties.get("marks", "set()"))

    @property
    def short_classname(self):
        return "/".join(self.classname.split(".")[1:-1]) + ".ini"

    @property
    def html_status(self):
        msg = []
        for status, nb in self._status.items():
            msg.append(f'<span class="alert-{status}">{status}</span>')
        return " ".join(msg)


class TestCase:
    def __init__(self, xml):
        for k, v in xml.attrib.items():
            setattr(self, k, v)
        self.properties = {}
        for prop in xml.findall("./properties/property"):
            self.properties[prop.attrib["name"]] = prop.attrib["value"]
        # ---------------------------------------------------------------
        # skipped
        skipped = xml.find("./skipped")
        if skipped is not None:
            self.skipped = skipped.attrib
        else:
            self.skipped = None
        # ---------------------------------------------------------------
        # failed
        failure = xml.find("./failure")
        if failure is not None:
            self.failure = failure.attrib
        else:
            self.failure = None

    @property
    def status(self):
        if self.skipped:
            return "skipped"
        elif self.failure:
            return "failed"
        return "success"

    def __repr__(self):
        return f"{self.classname}::{self.name}"


def parse_xml(filepath):
    """parse JUnit  XML file and return a tuple of `TestSuite` instances"""
    fpath = Path(filepath)
    assert fpath.exists()
    tree = ET.parse(filepath)
    root = tree.getroot()
    # get all testsuites
    return tuple((TestSuite(ts) for ts in root.findall("./testsuite")))


def xml2html(xmlfpath):
    # get mystranversion
    version, date = os.environ["MYSTRAN_VALIDATION_VERSION"].split("___")
    # =========================================================================
    # generate run HTML output
    # =========================================================================
    tss = parse_xml(xmlfpath)
    rtemplate = env.get_template("html_template.j2")
    kwargs = {"mystran_version": version, "mystran_date": date}
    kwargs["testsuites"] = tss
    return rtemplate.render(kwargs)
    # =========================================================================
    # generate index
    # =========================================================================
    itemplate = env.get_template("html_index.j2")
    rootdir = Path(os.environ["MYSTRAN_VALIDATION_ALLBUILDS"])

    with working_directory(rootdir):
        xmls = glob.glob("**/mystran-testing.xml", recursive=True)
    xmls = {os.path.split(xml)[0]: parse_xml(rootdir / xml)[0] for xml in xmls}
    matrix = defaultdict(dict)
    for version, ts in xmls.items():
        for ini_name, tcparent in ts.testcases.items():
            for tc in tcparent.testcases:
                matrix[version][
                    tc.__repr__().replace("mystran-test-cases.", "")
                ] = tc.status
    matrix = pd.DataFrame(dict(matrix))
    matrix.columns = [c.replace("___", " (") + ")" for c in matrix.columns]
    index = itemplate.render({"runs": list(xmls.keys()), "matrix": matrix})
    return index, rtemplate.render(kwargs)



import os
from contextlib import contextmanager
import re
import unicodedata
from functools import wraps
import asyncio
import subprocess
import shlex
import psutil

SUPERSCRIPT_REGEX = re.compile(r"\^([+|-]?\d+)")

MYSTRAN_VERSION_REGEX = re.compile(
    "MYSTRAN\s+[V|v]ersion\s+(?P<version>.*?)\s+(?P<date>(\w+\s+){3})"
)


def kill(proc_pid):
    process = psutil.Process(proc_pid)
    for proc in process.children(recursive=True):
        proc.kill()
    process.kill()


def get_mystran_version():
    cmd = os.environ["MYSTRAN_BIN"]
    proc = subprocess.Popen(
        shlex.split(cmd), stderr=subprocess.STDOUT, stdout=subprocess.PIPE
    )
    try:
        proc.wait(timeout=0.1)
    except subprocess.TimeoutExpired:
        kill(proc.pid)
    lines = proc.stdout.read().decode().split("\n")
    for line in lines:
        line = line.strip()
        if line.startswith("MYSTRAN Version"):
            match = MYSTRAN_VERSION_REGEX.match(line)
            version, date, *a = match.groups()
            mystran_attrs = {
                "version": version.strip(),
                "date": date.strip().replace(" ", "-"),
            }
            return mystran_attrs


def slugify(txt):

    """
    Slugify a unicode string.

    >>> slugify("Héllo Wörld-LCID#2")
    'hello_world_lcid2'
    >>> slugify("mystran-test-cases.BAR.no_offset.test_bar_06.ini")
    'bar_no_offset_test_bar_06_ini'
    """
    txt = unicodedata.normalize("NFKD", txt)
    txt = txt.replace("mystran-test-cases.", "")
    txt = txt.replace("mystran-test-cases" + os.sep, "")
    txt = txt.replace(os.sep, ".")
    txt = txt.replace(".", "_")
    txt = re.sub(r"[^\w\s-]", "", txt).strip().lower()
    return re.sub(r"[-\s]+", "_", txt)


@contextmanager
def working_directory(path):
    """
    A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.
    Usage:
    > # Do something in original directory
    > with working_directory('/my/new/path'):
    >     # Do something in new directory
    > # Back to old directory
    """

    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



# -*- coding: utf-8 -*-

"""Top-level package for MYSTRAN validation."""

__author__ = """Nicolas Cordier"""
__email__ = "nicolas.cordier@numeric-gmbh.ch"
__version__ = "0.13.0"

import os
import configparser
from pathlib import Path
import logging
import shutil
from functools import wraps
import asyncio
import datetime as dt
from appdirs import AppDirs

from .utils import slugify


def assert_frame_equal(dfa, dfb, atol, rtol):
    """it's hard to understand how pandas tesing is actually working
    better to fall back to numpy testing, assuming we do not need to
    check columns, index, etc...
    """
    # reorder columns
    dfb = dfb[[c for c in dfa if c in dfb]]
    diff = abs(dfa - dfb)
    crit = atol + rtol * abs(dfb)
    failing = diff[diff > crit]
    # keep rows and columns where at least one failure occurs
    failures = failing.dropna(how="all").dropna(how="all", axis=1)
    abs_error = failing.max().max()
    rel_error = (failing / abs(dfb)).max().max()
    failing = diff > crit
    return failing, failures, abs_error, rel_error


def background(func):
    """decorator for background jobs"""

    @wraps(func)
    def wrapped(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(None, func, *args, *kwargs)

    return wrapped


@background
def acopy(src, dest, **kwargs):
    logging.debug(f"{src}->{dest}")
    shutil.copy(src, dest, **kwargs)


def cleandir(target, **kwargs):
    if not isinstance(target, Path):
        target = Path(target)
    if target.exists():
        shutil.rmtree(target)
    target.mkdir(**kwargs)


def diffxlsx_target(name):
    dumping_dir = Path(os.environ["MYSTRAN_VALIDATION_BUILDDIR"]) / "dumps"
    dumping_dir.mkdir(exist_ok=True, parents=True)
    target = dumping_dir / (slugify(str(name)) + ".xlsx")
    return target


# =============================================================================
# CONFIGURATION stuff and default options
# =============================================================================
DIRS = AppDirs("mystran-validation", "numeric")

DEFAULTS = {
    "DEFAULT": {
        "mystran-bin": os.getenv("MYSTRAN_BIN", shutil.which("mystran")),
        "rootdir": os.path.join(DIRS.user_data_dir, "mystran-test-cases"),
        "report": "1",
        "open-report": "1",
        "builddir": "_build",
    }
}


def init_config(profile_name="DEFAULT", **kwargs):
    """create/update a profile in the config file. If config file not found,
    create it from scratch
    """
    config_fpath = Path(DIRS.user_config_dir) / "config.ini"
    parser = configparser.ConfigParser()
    params = DEFAULTS.copy()
    if profile_name not in params:
        params[profile_name] = {}
    for k, v in kwargs.items():
        if not v:
            continue
        params[profile_name][k.replace("_", "-")] = v
    parser.read_dict(params)
    config_fpath.parent.mkdir(parents=True, exist_ok=True)
    with open(config_fpath, "w") as configfile:
        parser.write(configfile)
    logging.debug(f"created {config_fpath}")
    return config_fpath


def is_comment(param):
    return param.strip()[0] in ("#", ";")


def get_conf():
    config_fpath = Path(DIRS.user_config_dir) / "config.ini"
    if not config_fpath.exists():
        raise FileNotFoundError(f"Configuration file '{str(config_fpath)}' not found")
        # init_config()
    parser = configparser.ConfigParser()
    parser.read(config_fpath)
    # -------------------------------------------------------------------------
    # always check if "DEFAULT" is up-to-date
    defaults = set((p for p in DEFAULTS["DEFAULT"].keys() if not is_comment(p)))
    actual = set((p for p in parser["DEFAULT"].keys() if not is_comment(p)))
    if defaults ^ actual:
        # some changes will occur, backup initial config
        backup_fpath = config_fpath.parent / (
            config_fpath.stem
            + ".ini.backup_"
            + dt.datetime.now().isoformat(timespec="minutes")
        )
        with open(backup_fpath, "w") as configfile:
            parser.write(configfile)
        logging.info(f"backuped {backup_fpath} before modification")
    # missing parameters
    for missing_param in defaults - actual:
        parser["DEFAULT"][missing_param] = DEFAULTS["DEFAULT"][missing_param]
        logging.info(f"added missing parameter {missing_param}")
    # unknown parameters
    for unknown_param in actual - defaults:
        _val = parser["DEFAULT"].pop(unknown_param)
        logging.warning(f"removed unknown parameter {unknown_param}={_val}")
    with open(config_fpath, "w") as configfile:
        parser.write(configfile)
    return config_fpath, parser


def get_profile(config, profile_name=None):
    if profile_name:
        try:
            config = dict(config[profile_name].items())
        except KeyError:
            msg = f"{profile_name} not defined in configuration file"
            raise KeyError(msg)
    else:
        config = dict(config["DEFAULT"].items())
    for k, v in config.items():
        config[k] = os.path.expanduser(v)
    return config



# content of conftest.pybreakpoint()
import ast
import configparser
import numbers
import os
import shlex
import shutil
import subprocess
import traceback
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from femap_neutral_parser import Parser as NeuParser

from mystran_validation import assert_frame_equal, cleandir, diffxlsx_target
from mystran_validation.parsers import subset
from mystran_validation.parsers.nastran_op2 import Parser as OP2Parser

# =============================================================================
# test collection
# =============================================================================


def pytest_collection_modifyitems(items):
    """add CARDs to marks"""
    for item in items:
        bulk_file = item.spec["bulk"]
        with open(bulk_file, "r") as fh:
            lines = fh.readlines()
        words = set(
            [
                line.split(" ")[0].strip()
                for line in lines
                if not line.startswith("$")
                and not line.startswith("+")
                and not line.startswith("PARAM,")
                and not line.startswith("CORD,")
            ]
        ) - set(
            [
                "",
                " ",
                "PARAM",
                "CEND",
                "ENDDATA",
                "NASTRAN",
                "ID",
                "GRID",
                "SUBCASE",
                "TIME",
                "BEGIN",
                "SOL",
                "INIT",
            ]
        )
        for tag in words:
            item.add_marker(getattr(pytest.mark, tag))
        # ---------------------------------------------------------------------
        # user-defined marks
        marks = item.spec.get("marks", ())
        if marks:
            marks = [m.strip() for m in marks.split(",")]
            for mark in marks:
                item.add_marker(getattr(pytest.mark, mark))


def pytest_collect_file(parent, path):
    _path = Path(path)
    if _path.suffix == ".ini" and _path.name.startswith("test"):
        return IniTestFile.from_parent(parent, fspath=path)


class IniTestFile(pytest.File):
    bulkfile = None
    rootdir = None
    rootname = None
    mystran_run_status = {}
    reference = {}
    op2s = {}

    def collect(self):
        config = configparser.ConfigParser()
        config.read(self.fspath)
        # ---------------------------------------------------------------------
        # prepare paths
        self.rootdir = Path(self.fspath).parent
        self.rootname = Path(self.fspath).stem
        # ---------------------------------------------------------------------
        # clean working dir
        wdir = self.rootdir / (".out_" + self.rootname)
        for name in config.sections():
            if name.startswith("_"):
                continue
            spec = dict(config[name].items())
            spec["test-config"] = Path(self.fspath)
            # =================================================================
            # building and yielding atomic test [name]
            # =================================================================
            bulkfile = self.rootdir / spec["bulk"]
            if self.bulkfile:
                if bulkfile != self.bulkfile:
                    raise ValueError(f"{name}: One class, one bulk, one output!")
            else:
                self.bulkfile = bulkfile
            spec["workingdir"] = wdir
            spec["bulk"] = bulkfile
            ref = spec["reference"]
            # -----------------------------------------------------------------
            # if ref is obviously a file
            if Path(ref).suffix.lower() in (".op2", ".neu"):
                spec["reference"] = self.rootdir / spec["reference"]
            # -----------------------------------------------------------------
            # if ref a sing value
            else:
                ref = float(ref)
                spec["reference"] = ref
            spec["rtol"] = float(spec.get("rtol", 1e-05))
            spec["atol"] = float(spec.get("atol", 1e-08))
            spec["shortdesc"] = spec.get("shortdesc", "")
            spec["output"] = spec["workingdir"] / (bulkfile.stem + ".NEU")
            spec["output_f06"] = spec["workingdir"] / (bulkfile.stem + ".F06")
            spec["diffxlsx"] = spec["workingdir"] / (bulkfile.stem + ".NEU")
            yield IniItem.from_parent(self, name=name, spec=spec)


class IniItem(pytest.Item):
    """Atomic test. This is where one single [test] is performed"""

    def __init__(self, name, parent, spec):
        super().__init__(name, parent)
        self.spec = spec

    def run_mystran(self):
        """run bulkfile"""
        wdir = self.spec["workingdir"]
        cleandir(wdir)
        target = shutil.copy(self.spec["bulk"], self.spec["workingdir"])
        cmd = f"{os.getenv('MYSTRAN_BIN')} {target}"
        status = subprocess.check_output(
            shlex.split(cmd),
            stderr=subprocess.STDOUT,
        )
        if "fatal" in status.decode().lower():
            # retrieve Error
            target = Path(target)
            error_file = target.parent / (target.stem + ".ERR")
            # f06_file = target.parent / (target.stem + ".F06")
            with open(error_file, "r") as fh:
                lines = fh.readlines()
            errors = []
            _parsing_error = False
            for line in lines:
                line = line.strip()
                if _parsing_error:
                    if line.startswith("*"):
                        errors.append(tuple(_parsing_error))
                        _parsing_error = False
                    else:
                        _parsing_error.append(line)
                if line.startswith("*ERROR"):
                    _parsing_error = [line]
            if _parsing_error:
                errors.append(tuple(_parsing_error))
            errors = ["\n".join(err) for err in errors]
            raise MystranException(errors[0])
        # ---------------------------------------------------------------------
        # get actual results
        neu = NeuParser(self.spec["output"])
        self.parent.actual = neu
        self.parent.actual.info(doprint=False)  # pre-digest data
        self.parent.actual_available_vectors = sorted(
            self.parent.actual._output_vectors.keys()
        )
        return status

    def expected(self, vector, filters=None, axis=None, raw=False):
        """build reference (expected) DataFrame"""
        if not filters:
            filters = {}
        df_expected = None
        # ---------------------------------------------------------------------
        # reference is a numerical value
        if isinstance(self.spec["reference"], numbers.Number):
            if axis is None:
                raise ValueError("param `axis` must be specified for manual checking")
            _data = filters.copy()
            _data.update({axis: [self.spec["reference"]]})
            df_expected = pd.DataFrame(_data)
            df_expected.set_index(list(filters.keys()), inplace=True)
            df_expected = df_expected[axis].to_frame()
        # ---------------------------------------------------------------------
        # reference is a ref file
        elif self.spec["reference"].suffix.lower() == ".op2":
            op2 = self.parent.op2s[self.spec["reference"]]
            df_expected = op2.get(vector=vector, raw=raw, **filters)
        else:
            raise ValueError("reference {self.spec['reference']} not understood")
        return df_expected

    def actual(self, vector, filters=None, axis=None):
        """build actual DataFrame"""
        if not filters:
            filters = {}
        # ---------------------------------------------------------------------
        # as of femap_neutral_parser 0.8, vectos have been renamed
        compat = {
            "reactions": "spc_forces",
        }
        if vector in compat:
            warnings.warn(
                f"``{vector}`` is deprecated. Use ``{compat[vector]}`` instead"
            )
        vector = compat.get(vector, vector)
        # -----------------------------------------------------------------
        # get values from neutral file, and reshape it
        df_actual = self.parent.actual.get(vector, asdf=True)
        df_actual = subset(df_actual, **filters)
        if axis:
            try:
                df_actual = df_actual[axis].to_frame()
            except KeyError:
                msg = f"{axis} is not a proper column. Use one of {df_actual.columns.tolist()}"
                raise KeyError(msg)
        return df_actual

    def runtest(self):
        """trigger pytest"""
        self.user_properties += [("bulk", str(self.spec["bulk"]))]
        self.user_properties += [("ref", str(self.spec["reference"]))]
        self.user_properties += [("description", self.spec["description"])]
        self.user_properties += [("atol", self.spec["atol"])]
        self.user_properties += [("rtol", self.spec["rtol"])]
        self.user_properties += [("shortdesc", self.spec["shortdesc"])]
        self.user_properties += [("vector", self.spec["vector"])]
        self.user_properties += [("marks", set([m.name for m in self.own_markers]))]
        self.user_properties += [("test-config", self.spec["test-config"])]
        if self.spec.get("skip"):
            pytest.skip(self.spec["skip"])
        # ---------------------------------------------------------------------
        # run Mystran only once
        if self.spec["bulk"] not in self.parent.mystran_run_status:
            # first test for this class, a few things to set up
            retmystran = self.run_mystran()
            self.parent.mystran_run_status[self.spec["bulk"]] = retmystran
        # process OP2 only once
        ref = self.spec["reference"]
        if not isinstance(ref, numbers.Number):
            if ref.suffix.lower() == ".op2" and str(ref) not in self.parent.op2s:
                self.parent.op2s[ref] = OP2Parser(str(self.spec["reference"]))
            self.op2 = self.parent.op2s[ref]  # shortcut for self
        else:
            self.op2 = None
        self.user_properties += [("output", str(self.spec["output"]))]
        self.user_properties += [("output_f06", str(self.spec["output_f06"]))]
        try:
            vector = self.spec["vector"]
            # ---------------------------------------------------------------------
            # get subset index
            filters = dict(
                SubcaseID=ast.literal_eval(self.spec.get("subcaseids", "None")),
                NodeID=ast.literal_eval(self.spec.get("nodeids", "None")),
                ElementID=ast.literal_eval(self.spec.get("elementids", "None")),
            )
            axis = self.spec.get("axis")
            filters = {k: [v] for k, v in filters.items() if v}
            # ---------------------------------------------------------------------
            # get dataframes to compare
            df_actual = self.actual(vector, filters, axis)
            df_expected = self.expected(vector, filters, axis)
            # -----------------------------------------------------------------
            # check tolerances
            rtol = self.spec["rtol"]
            atol = self.spec["atol"]
            failing, failures, aerr, rerr = assert_frame_equal(
                df_actual, df_expected, rtol=rtol, atol=atol
            )
        except Exception as exc:
            tb = traceback.format_exc()
            raise GenericException(self, exc, tb)
        else:
            if len(failures) > 0:
                # dump comparisons performed
                if len(failures) > 0:
                    xlsxdiff_target = diffxlsx_target(self.parent.name)
                    xlsx_filepath = dump(
                        df_expected,
                        df_actual,
                        target=xlsxdiff_target,
                        sheetname=self.name,
                        failing=failing,
                    )
                    self.user_properties += [("xlsxdiff", xlsxdiff_target)]
                raise IniException(
                    self, df_actual, df_expected, failures, rtol, atol, aerr, rerr
                )

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        if isinstance(excinfo.value, MystranException):
            return excinfo.value
        elif isinstance(excinfo.value, IniException):
            (
                item,
                df_actual,
                df_expected,
                failures,
                rtol,
                atol,
                aerr,
                rerr,
            ) = excinfo.value.args
            df_actual = df_actual.loc[failures.index]
            df_expected = df_expected.loc[failures.index]
            # failing difference
            return "\n".join(
                [
                    f"usecase `{self.fspath}::[{self.name}]`\nexecution failed given precision requirements:\n  * {atol=}\n  * {rtol=}\n",
                    f"failing with:\n  * Absolute difference {aerr=}\n  * Relative difference {rerr=}\n",
                    f"Expected\n--------\n{df_expected}\n",
                    f"Actual\n------\n{df_actual}",
                ]
            )
        elif isinstance(excinfo.value, GenericException):
            item, exc, traceback = excinfo.value.args
            return "\n".join(
                [
                    f"usecase `{self.fspath}::[{self.name}]` raised the following exception\n",
                    f"{exc}\n",
                    f"file: {traceback}",  # get rid of "file " prefix
                ]
            )

    def reportinfo(self):
        return self.fspath, 0, f"usecase: {self.name}"


class IniException(Exception):
    pass


class MystranException(Exception):
    pass


class GenericException(Exception):
    pass


def apply_color(x, failing):
    # colors = {False: "green", True: "red; font-weight: bold", "ref": "black"}
    colors = {
        False: "green",
        True: "red; text-decoration: line-through;",
        "ref": "black",
    }
    return failing.applymap(lambda val: "color: {}".format(colors.get(val, "")))


def highlight_columns(x):
    if x.name[1] == "mystran":
        style = "background-color: azure"
    else:
        style = "background-color: white"
    return [style] * x.shape[0]


def dump(df_expected, df_actual, target, sheetname, failing, debug=False):
    full = pd.concat({"ref.": df_expected, "mystran": df_actual}).unstack(level=0)
    # reshape failing to have same format as `full`
    _ref = pd.DataFrame(
        data=np.full_like(failing, "ref", dtype=object),
        index=failing.index,
        columns=failing.columns,
    )
    _failing = pd.concat({"ref.": _ref, "mystran": failing}).unstack(level=0)
    # ensure columns are ordered as `full` DF
    _failing = _failing[[c for c in full.columns]]
    # =========================================================================
    # formatting
    # =========================================================================
    mode = "a" if target.exists() else "w"
    # style dataframe
    styled = full.style.apply(apply_color, axis=None, failing=_failing)
    styled = styled.apply(highlight_columns)
    with pd.ExcelWriter(target, mode=mode, engine="openpyxl") as writer:
        styled.to_excel(writer, sheet_name=sheetname)
    return target
